package ru.landgrafhomyak.db.sql_skeleton_0.api.queries

import kotlin.jvm.JvmField
import kotlin.jvm.JvmName
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.Column
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.IntermediateColumn
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.DatabaseType
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.Mapper
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.FiltersScope
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.QueryParam
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc.Reducer
import ru.landgrafhomyak.db.sql_skeleton_0.api.misc._Selectable
import ru.landgrafhomyak.db.sql_skeleton_0.api.runtime.InputRow

public interface SelectQuery<out qUE : Any> : _Query<qUE> {
	public enum class Order {
		ASC, DESC
	}

	public abstract class JoinType {
		public class LeftJoin(
			@JvmField
			public val leftNull: Null,
			@JvmField
			public val behaviour: Behaviour
		) : JoinType() {
			@Suppress("ClassName")
			public abstract class Null {
				public object NULL_NOT_EXPECTED : Null()
				public object AS_ID : Null()
				public object SKIP : Null()
				public object JOIN_TO_NULL : Null()
			}

			@Suppress("ClassName")
			public abstract class Behaviour {
				public object LEFT_TO_UNIQUE_RIGHT : Behaviour()
				public object LEFT_TO_UNIQUE_RIGHT_OR_TO_NULL : Behaviour()
				public object LEFT_FOR_EACH_RIGHT : Behaviour()
				public object LEFT_FOR_EACH_RIGHT_OR_TO_NULL : Behaviour()
			}
		}

		public object CrossJoin : JoinType()
	}

	@Suppress("ClassName")
	public interface _CommonSelectCreatorScope<qUE : Any> {
		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("types")
		public val types: DatabaseType.Provider
		public fun <RT, DT : DatabaseType<RT>> queryParam(name: String, type: DT): QueryParam<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> paramAsColumn(param: QueryParam<RT, DT, qUE>): IntermediateColumn<RT, DT, qUE>

		public fun <mUE : Any> createMapper(constructor: Mapper.Constructor<mUE>): Mapper<mUE>
		public fun <mUE : Any, oUE : Any> mapColumns(
			mapper: Mapper<mUE>,
			input: Mapper.InputLinker<mUE, qUE>,
			output: Mapper.OutputLinker<mUE, qUE, oUE>
		): oUE

		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("filters")
		public val filters: FiltersScope<qUE>
		public fun orderBy(order: Order, vararg column: IntermediateColumn<*, *, qUE>)

		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("_getLimit")
		@set:JvmName("limit")
		public var limit: UInt

		@Suppress("INAPPLICABLE_JVM_NAME")
		@get:JvmName("_getOffset")
		@set:JvmName("offset")
		public var offset: ULong

		public fun <RT, DT : DatabaseType<RT>> returnParam(name: String, param: QueryParam<RT, DT, qUE>): Column<RT, DT, qUE>
		public fun <RT, DT : DatabaseType<RT>> returnColumn(name: String, column: IntermediateColumn<RT, DT, qUE>): Column<RT, DT, qUE>
	}

	public interface Simple<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Simple<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Simple<qUE> =
				outQueries.select(this)

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("source")
			public val source: _Selectable<tUE>

			public fun createSelect(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("src")
				public val src: _Selectable<tUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("subqueryParams")
				public val subqueryParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>
			}
		}
	}

	public interface WithJoin<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, lsqUE : Any, rsqUE : Any> : _Query.Constructor<WithJoin<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): WithJoin<qUE> =
				outQueries.selectWithJoin(this)

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("left")
			public val left: _Selectable<lsqUE>

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("right")
			public val right: _Selectable<rsqUE>

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("joinType")
			public val joinType: JoinType

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("leftColumnToJoin")
			public val leftColumnToJoin: Column<*, *, lsqUE>

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("rightColumnToJoin")
			public val rightColumnToJoin: Column<*, *, rsqUE>

			public fun createSelectWithJoin(context: Scope<qUE, lsqUE, rsqUE>): qUE

			public interface Scope<qUE : Any, lsqUE : Any, rsqUE : Any> : _CommonSelectCreatorScope<qUE> {
				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("lsq")
				public val lsq: _Selectable<lsqUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("rsq")
				public val rsq: _Selectable<rsqUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("leftSubqueryParams")
				public val leftSubqueryParams: InputRow.WithRedirect<lsqUE, qUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("rightSubqueryParams")
				public val rightSubqueryParams: InputRow.WithRedirect<rsqUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromLeft(column: Column<RT, DT, lsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT, DT : DatabaseType<RT>> selectColumnFromRight(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT, DT, qUE>
				public fun <RT : Any, DT : DatabaseType<RT>> selectColumnFromRightAsNullable(column: Column<RT, DT, rsqUE>): IntermediateColumn<RT?, DatabaseType.Nullable<RT, DT>, qUE>
			}
		}
	}

	public interface Reducing<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Reducing<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Reducing<qUE> =
				outQueries.selectReducing(this)

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("source")
			public val source: _Selectable<tUE>

			public fun createSelectWithReduce(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("src")
				public val src: _Selectable<tUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("sourceParams")
				public val sourceParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>

				public fun <rUE : Any> createReducer(constructor: Reducer.Constructor<rUE>): Reducer<rUE>
				public fun <mUE : Any, oUE : Any> reduceRows(
					reducer: Reducer<mUE>,
					linker: Reducer.Linker<mUE, qUE, oUE>,
				): oUE

				public fun groupBy(vararg columns: IntermediateColumn<*, *, qUE>)
			}
		}
	}

	public interface Folding<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, tUE : Any> : _Query.Constructor<Folding<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Folding<qUE> =
				outQueries.selectFolding(this)

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("source")
			public val source: _Selectable<tUE>

			public fun createSelectWithFold(context: Scope<qUE, tUE>): qUE

			public interface Scope<qUE : Any, tUE : Any> : _CommonSelectCreatorScope<qUE> {
				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("src")
				public val src: _Selectable<tUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("sourceParams")
				public val sourceParams: InputRow.WithRedirect<tUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> selectColumnFromSubquery(param: Column<RT, DT, tUE>): IntermediateColumn<RT, DT, qUE>

				public fun <rUE : Any> createReducer(constructor: Reducer.Constructor<rUE>): Reducer<rUE>
				public fun <mUE : Any, oUE : Any> foldRows(
					reducer: Reducer<mUE>,
					linker: Reducer.Linker<mUE, qUE, oUE>,
				): oUE

				public fun groupBy(vararg columns: IntermediateColumn<*, *, qUE>)
			}
		}
	}


	public interface Union<out qUE : Any> : SelectQuery<qUE>, _Query.Params2Table<qUE> {
		public interface Constructor<qUE : Any, fqUE : Any, lqUE : Any> : _Query.Constructor<Union<qUE>> {
			override fun createQuery(internalQueries: _Query.Constructor.Scope, outQueries: _Query.Constructor.Scope): Union<qUE> =
				outQueries.selectUnion(this)

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("first")
			public val first: _Selectable<fqUE>

			@Suppress("INAPPLICABLE_JVM_NAME")
			@get:JvmName("last")
			public val last: _Selectable<lqUE>

			public fun createSelectWithJoin(context: Scope<qUE, fqUE, lqUE>): qUE

			public interface Scope<qUE : Any, fqUE : Any, lqUE : Any> : _CommonSelectCreatorScope<qUE> {
				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("fq")
				public val fq: _Selectable<fqUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("lq")
				public val lq: _Selectable<lqUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("firstSubqueryParams")
				public val firstSubqueryParams: InputRow.WithRedirect<fqUE, qUE>

				@Suppress("INAPPLICABLE_JVM_NAME")
				@get:JvmName("lastSubqueryParams")
				public val lastSubqueryParams: InputRow.WithRedirect<lqUE, qUE>

				public fun <RT, DT : DatabaseType<RT>> uniteColumns(first: Column<RT, DT, fqUE>, last: Column<RT, DT, lqUE>): IntermediateColumn<RT, DT, qUE>
			}
		}
	}
}